SystemOrganization addCategory: #'Smallapack-OpalCompiler'!

OCASTTranslatorForValue subclass: #SLASTTranslatorForValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLASTTranslatorForValue commentStamp: 'nice 1/9/2019 21:12' prior: 0!
I am an auxiliary of AST to IR translator for Smallapack Opal Compiler.

I generate instructions that keep resulting value on the execution  stack.

I am specialized for sending  messages with more than 15 arguments.!

!SLASTTranslatorForValue methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:42'!
classForEffect
	^SLASTTranslatorForEffect ! !

!SLASTTranslatorForValue methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:44'!
classForValue
	^SLASTTranslatorForValue! !

!SLASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 18:25'!
emitMessageNode: aMessageNode
	aMessageNode arguments size < 16 ifTrue: [ ^super emitMessageNode: aMessageNode ].
	^ self emitMessageNodeWithManyArgs: aMessageNode! !

IRBytecodeGenerator subclass: #SLIRBytecodeGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLIRBytecodeGenerator commentStamp: 'nice 1/9/2019 21:34' prior: 0!
I transform IR instructions into a sequence of bytecodes for Smallapack Opal Compiler.

I am specialized for sending messages with more than 15 arguments.
I also handle a correct method header for a method with more than 15 arguments.!

!SLIRBytecodeGenerator methodsFor: 'accessing' stamp: 'nice 1/9/2019 21:33'!
numArgs: anInteger
	"For writing the CompiledMethod header, we know that 16 arguments or more will be passed into a single array argument."
	numArgs := anInteger > 15
		ifTrue: [ 1 ]
		ifFalse: [ anInteger ]! !

!SLIRBytecodeGenerator methodsFor: 'instructions' stamp: 'nice 1/9/2019 14:40'!
send: selector
	| nArgs |
	nArgs := selector numArgs.
	nArgs > 15 ifTrue: [ nArgs := 1 ].
	stack pop: nArgs.
	(self encoderClass specialSelectors includes: selector)
		ifTrue: [ ^ encoder genSendSpecial: (self encoderClass specialSelectors indexOf: selector) numArgs: nArgs ].
	encoder genSend: (self literalIndexOf: selector) numArgs: nArgs! !

!SLIRBytecodeGenerator methodsFor: 'instructions' stamp: 'nice 1/9/2019 14:38'!
send: selector toSuperOf: behavior

	(encoder class = OpalEncoderForSistaV1 and: [ inBlock ])
		ifTrue: [ 
			| index nArgs |
			behavior isTrait ifTrue: [ 
				"Trait methods are copied to the users and only the last literal is updated. 
				 For directed super send the literal of the super send should be updated too."
				self error: 'not supported' ].
			nArgs := selector numArgs.
			nArgs > 15 ifTrue: [ nArgs := 1 ].
			stack pop: nArgs.
			encoder genPushLiteralVar: (self literalIndexOf: behavior binding).
			index := self literalIndexOf: selector.
			encoder genSendDirectedSuper: index numArgs: nArgs ]
		ifFalse: [ 
			| index nArgs |
			nArgs := selector numArgs.
			nArgs > 15 ifTrue: [ nArgs := 1 ].
			stack pop: nArgs.
			self addLastLiteral: behavior binding.  
			index := self literalIndexOf: selector.
			encoder genSendSuper: index numArgs: nArgs ]
! !

!IRPrinterV2 methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:31'!
visitPushArrayedArg: pushArrayedArg

	stream nextPutAll: 'pushArrayedArg: '.
	pushArrayedArg name printOn: stream..
	stream nextPutAll: ' inVector: '.
	pushArrayedArg tempVectorName printOn: stream! !

!IRInstruction class methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:52'!
create1arrayForAllArgumentNamed: aTempVectorName withVars: anArray  

	^ IR1ArrayForAllArgument new
		name: aTempVectorName;
		vars: anArray;
		yourself.! !

!IRInstruction class methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:46'!
pushArrayedArg: aName inVectorAt: nameOfVector
	^ IRPushArrayedArg new
		name: aName;
		tempVectorName: nameOfVector;
		yourself.! !

IRPushRemoteTemp subclass: #IRPushArrayedArg
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!IRPushArrayedArg commentStamp: 'nice 1/9/2019 20:57' prior: 0!
I represent an instruction for pushing  an argument of a method on the stack.

I am used when arguments have been packed into a single array.!

!IRPushArrayedArg methodsFor: 'visiting' stamp: 'nice 1/9/2019 18:29'!
accept: aVisitor
	^ aVisitor visitPushArrayedArg: self! !

!OCAbstractMethodScope methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 17:03'!
arrayArgumentName
	^'1arrayForAllArguments'! !

!IRVisitor methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:42'!
visit1ArrayForAllArgument: oneArray! !

!IRVisitor methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:30'!
visitPushArrayedArg: pushArrayedArg! !

!IRBuilder methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:52'!
create1arrayForAllArgumentNamed: name withVars: anArray
	
	"self addVectorTemps: anArray."
	self addTemp: name.
	self add: (IRInstruction create1arrayForAllArgumentNamed: name withVars: anArray).! !

!IRBuilder methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:47'!
pushArrayedArg: name inVector: nameOfVector

	^self add: (IRInstruction pushArrayedArg: name inVectorAt: nameOfVector)! !

IRBuilder subclass: #SLIRBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLIRBuilder commentStamp: 'nice 1/9/2019 20:51' prior: 0!
I am the IRBuilder for Smallapack Opal Compiler

My sole interest is to use SLIRMethod!

!SLIRBuilder methodsFor: 'initialization' stamp: 'nice 1/9/2019 15:03'!
initialize
	ir := SLIRMethod new.
	jumpAheadStacks := IdentityDictionary new.
	jumpBackTargetStacks := IdentityDictionary new.
	sourceMapNodes := OrderedCollection new.	"stack"
	currentScope := Stack new.
	self pushScope: ir.
	
	"Leave an empty sequence up front (guaranteed not to be in loop)"
	ir startSequence: ((IRSequence orderNumber: 0) method:ir).
	currentSequence := (IRSequence orderNumber: 1) method:ir.
	ir startSequence add:
		(IRJump new
			destination: currentSequence;
			bytecodeIndex: sourceMapByteIndex;
			yourself).
! !

RFASTTranslatorForValue subclass: #SLRFASTTranslatorForValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLRFASTTranslatorForValue commentStamp: 'nice 1/9/2019 21:13' prior: 0!
I am an auxiliary of Reflective AST to IR translator for Smallapack Opal Compiler.

I generate instructions that keep resulting value on the execution  stack.

I am specialized for sending  messages with more than 15 arguments.!

!SLRFASTTranslatorForValue methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:43'!
classForEffect
	^SLRFASTTranslatorForEffect ! !

!SLRFASTTranslatorForValue methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:44'!
classForValue
	^SLRFASTTranslatorForValue! !

!SLRFASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:31'!
emitMessageNode: aMessageNode
	aMessageNode arguments size < 16 ifTrue: [ ^super emitMessageNode: aMessageNode ].
	^ self emitMessageNodeWithManyArgs: aMessageNode! !

ReflectiveMethod variableSubclass: #SLReflectiveMethod
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLReflectiveMethod commentStamp: 'nice 1/9/2019 21:28' prior: 0!
I am a Reflective method for Smallapack Opal compiler.

I am using specialized ancillary classes that can deal with messages/methods with more than 15 args.

I am currently unused, because the class name for original Opal Compiler is encoded to deeply in existing code base (CompiledMethod), see references to my superclass ReflectiveMethod. CompiledMethod is a class known to the VM, so we cannot easily substitute it for referencing SLReflectiveMethod instead.

PLEASE FIX ME.!

!SLReflectiveMethod methodsFor: 'evaluation' stamp: 'nice 1/9/2019 21:24'!
generatePrimitiveWrapper
	| wrappedMethod send wrapperMethod |
	OCASTSemanticCleaner clean: ast.
	ast compilationContext 
		semanticAnalyzerClass: SLRFSemanticAnalyzer;
		astTranslatorClass: SLRFASTTranslator.
	ast doSemanticAnalysis. "force semantic analysis"
	wrappedMethod := ast generate: compiledMethod trailer.
	
	send := RBMessageNode
		receiver: (RBSelfNode named: #self)
		selector:  #rFwithArgs:executeMethod:
		arguments: {RBArrayNode statements: ast arguments . (RBLiteralNode value: wrappedMethod)}.
	
	wrapperMethod := RBMethodNode
		selector: ast selector
		arguments: ast arguments 
		body: (RBReturnNode value: send) asSequenceNode.
		
	wrapperMethod methodClass: ast methodClass.
	wrapperMethod propertyAt: #wrappedPrimitive put: true.
	ast hasMetalink ifTrue: [wrapperMethod propertyAt: #links put: (ast propertyAt: #links)].
	ast := wrapperMethod.! !

!SLReflectiveMethod methodsFor: 'evaluation' stamp: 'nice 1/9/2019 15:14'!
recompileAST
	| links |
	OCASTSemanticCleaner clean: ast.
	ast compilationContext 
		semanticAnalyzerClass: RFSemanticAnalyzer;
		astTranslatorClass: SLRFASTTranslator.
	ast doSemanticAnalysis. "force semantic analysis"
	compiledMethod := ast generate: compiledMethod trailer.
	compiledMethod reflectiveMethod: self.! !

!RFASTTranslator methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 17:30'!
emitMessageNodeWithManyArgs: aMessageNode
	| argArrayNode |
	self emitPrepareLinkAfter: aMessageNode.
	aMessageNode isCascaded
		ifFalse: [ valueTranslator visitNode: aMessageNode receiver ].
	argArrayNode := (RBArrayNode statements: aMessageNode arguments).
	valueTranslator visitNode: argArrayNode.
	self emitMetaLinkBefore: aMessageNode.
	aMessageNode hasMetalinkInstead
		ifTrue: [ self emitMetaLinkInstead: aMessageNode ]
		ifFalse: [ 
			aMessageNode isSuperSend
				ifTrue: [ methodBuilder send: aMessageNode selector toSuperOf: self compilationContext getClass ]
				ifFalse: [ methodBuilder send: aMessageNode selector ] ].
	self emitMetaLinkAfter: aMessageNode.! !

!RFASTTranslator methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 22:16'!
visitMethodNodeWithManyArgs: aMethodNode
	"I ignore all links when I am primitive as ReflectiveMethod compiles a wrapper"
	
	methodBuilder compilationContext: aMethodNode compilationContext.
	methodBuilder addTemps: aMethodNode scope tempVarNames.
	
	aMethodNode isPrimitive ifFalse: [self emitMetaLinkBefore: aMethodNode].

	methodBuilder properties: aMethodNode properties.
	methodBuilder irPrimitive: aMethodNode primitiveFromPragma.
	aMethodNode pragmas do: [:each | self visitPragmaNode: each].
	methodBuilder numArgs: 1 "dilemna: answer the effective or the apparent number of arguments? The apparent would be: aMethodNode arguments size".
	
	methodBuilder 
		create1arrayForAllArgumentNamed: aMethodNode scope arrayArgumentName 
		withVars: (aMethodNode arguments collect: [:each| each name] as: Array).
	
	self emitPrepareLinkAfter: aMethodNode.
	
	(aMethodNode hasMetalinkInstead and: [ aMethodNode isPrimitive not ])
				ifTrue: [ self emitMetaLinkInstead: aMethodNode. methodBuilder returnTop.
						^self. ].
				
	aMethodNode scope tempVector ifNotEmpty: [
		methodBuilder 
			createTempVectorNamed: aMethodNode scope tempVectorName 
			withVars: (aMethodNode scope tempVector collect: [:each| each name]) asArray.
	].
	effectTranslator visitNode: aMethodNode body.
	aMethodNode body lastIsReturn ifFalse:  [methodBuilder pushReceiver; returnTop].
	self emitMetaLinkAfter: aMethodNode.
	aMethodNode hasMetalinkAfter ifTrue: [methodBuilder returnTop].
! !

RFASTTranslator subclass: #SLRFASTTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLRFASTTranslator commentStamp: 'nice 1/9/2019 21:05' prior: 0!
I am a reflective AST to IR translator for Smallapack Opal Compiler.

I know how to produce instructions for methods with more than 15 arguments.
And I use instruction generators for value/effect that know how to generate message sends with more than 15 arguments.!

!SLRFASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:43'!
classForEffect
	^SLRFASTTranslatorForEffect ! !

!SLRFASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:44'!
classForValue
	^SLRFASTTranslatorForValue! !

!SLRFASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 18:04'!
initialize

	methodBuilder := SLIRBuilder new.
	effectTranslator := self as: self classForEffect.
	valueTranslator := self as: self classForValue.
	effectTranslator instVarNamed: #effectTranslator put: effectTranslator.
	effectTranslator instVarNamed: #valueTranslator put: valueTranslator.
	valueTranslator instVarNamed: #valueTranslator put: valueTranslator.
! !

!SLRFASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:23'!
visitMethodNode: aMethodNode
	aMethodNode arguments size > 15 ifFalse: [ ^super visitMethodNode: aMethodNode ].
	^self visitMethodNodeWithManyArgs: aMethodNode! !

OCArgumentVariable variableSubclass: #SLArrayedArgumentVariable
	instanceVariableNames: 'vectorName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLArrayedArgumentVariable commentStamp: 'nice 1/9/2019 21:07' prior: 0!
I represent an argument variable for the case when several  arguments are passed into an array.

Such construct is used to bypass the 15-arguments limit.

I remember the name of the array variable (vectorName), and I know how to generate instructions for reading the argument variable (push it on the stack).

!

!SLArrayedArgumentVariable methodsFor: 'emitting' stamp: 'nice 1/9/2019 18:48'!
emitValue: methodBuilder

	methodBuilder pushArrayedArg: name inVector: vectorName.! !

!SLArrayedArgumentVariable methodsFor: 'accessing' stamp: 'nice 1/9/2019 16:31'!
vectorName
	^ vectorName! !

!SLArrayedArgumentVariable methodsFor: 'accessing' stamp: 'nice 1/9/2019 16:32'!
vectorName: anObject
	vectorName := anObject! !

!OCASTTranslator methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 17:27'!
emitMessageNodeWithManyArgs: aMessageNode
	"For sending a message with more than 15 arguments,
	stack all the arguments into a single array, and send with 1 array argument."
	| argArrayNode |
	argArrayNode := (RBArrayNode statements: aMessageNode arguments).
	aMessageNode isCascaded ifFalse: [
		valueTranslator visitNode: aMessageNode receiver].
	valueTranslator visitNode: argArrayNode.
	aMessageNode isSuperSend
		ifTrue: [methodBuilder send: aMessageNode selector toSuperOf: self compilationContext getClass]
		ifFalse: [methodBuilder send: aMessageNode selector].! !

!OCASTTranslator methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 21:36'!
visitMethodNodeWithManyArgs: aMethodNode
	methodBuilder compilationContext: aMethodNode compilationContext.
	methodBuilder addTemps: aMethodNode scope tempVarNames.

	methodBuilder properties: aMethodNode properties.
	methodBuilder irPrimitive: aMethodNode primitiveFromPragma.
	aMethodNode pragmas do: [:each | self visitPragmaNode: each].
	methodBuilder numArgs: 1 "dilemna: answer the effective or the apparent number of arguments? The apparent would be: aMethodNode arguments size".
	
	methodBuilder 
		create1arrayForAllArgumentNamed: aMethodNode scope arrayArgumentName 
		withVars: (aMethodNode arguments collect: [:each| each name] as: Array).

	aMethodNode scope tempVector ifNotEmpty: [
		methodBuilder 
			createTempVectorNamed: aMethodNode scope tempVectorName 
			withVars: (aMethodNode scope tempVector collect: [:each| each name]) asArray.
	].
	effectTranslator visitNode: aMethodNode body.
	aMethodNode body lastIsReturn ifFalse:  [methodBuilder pushReceiver; returnTop]! !

OCASTTranslator subclass: #SLASTTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLASTTranslator commentStamp: 'nice 1/9/2019 21:04' prior: 0!
I am an AST to IR translator for Smallapack Opal Compiler.

I know how to produce instructions for methods with more than 15 arguments.
And I use instruction generators for value/effect that know how to generate message sends with more than 15 arguments.!

!SLASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:42'!
classForEffect
	^SLASTTranslatorForEffect ! !

!SLASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:43'!
classForValue
	^SLASTTranslatorForValue! !

!SLASTTranslator methodsFor: 'initialization' stamp: 'nice 1/9/2019 18:04'!
initialize

	methodBuilder := SLIRBuilder new.
	effectTranslator := self as: self classForEffect.
	valueTranslator := self as: self classForValue.
	effectTranslator instVarNamed: #effectTranslator put: effectTranslator.
	effectTranslator instVarNamed: #valueTranslator put: valueTranslator.
	valueTranslator instVarNamed: #valueTranslator put: valueTranslator.
! !

!SLASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:23'!
visitMethodNode: aMethodNode
	aMethodNode arguments size > 15 ifFalse: [ ^super visitMethodNode: aMethodNode ].
	^self visitMethodNodeWithManyArgs: aMethodNode! !

IRMethod subclass: #SLIRMethod
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLIRMethod commentStamp: 'nice 1/9/2019 20:52' prior: 0!
I am the IRMethod for Smallapack Opal Compiler

My sole interest is to use SLCompilationContext!

!SLIRMethod methodsFor: 'accessing' stamp: 'nice 1/9/2019 15:01'!
compilationContext
	^ compilationContext ifNil: [ 
		"only happens when decompiling or using stand-alone"
		compilationContext := SLCompilationContext default]! !

!OCASTSemanticAnalyzer methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:23'!
declareArrayedArgumentNode: aVariableNode
	^self declareVariableNode: aVariableNode as: (SLArrayedArgumentVariable new vectorName: scope arrayArgumentName)! !

!OCASTSemanticAnalyzer methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:23'!
visitMethodNodeWithManyArgs: aMethodNode
	scope := compilationContext scope newMethodScope. 
	aMethodNode scope: scope.  scope node: aMethodNode.
	self declareArgumentNode: (RBArgumentNode named: scope arrayArgumentName).
	aMethodNode arguments do: [:node | self declareArrayedArgumentNode: node ].
	aMethodNode pragmas do: [:each | self visitNode: each].
	self visitNode: aMethodNode body.
	scope := scope outerScope.! !

OCASTSemanticAnalyzer subclass: #SLASTSemanticAnalyzer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLASTSemanticAnalyzer commentStamp: 'nice 1/9/2019 21:01' prior: 0!
I am the AST semantic analyzer for Smallapack Opal Compiler.

I am here for mapping argument names to argument index when all arguments are passed in a single array.

This single array argument construct is used for by-passing the 15-argument limit.!

!SLASTSemanticAnalyzer methodsFor: 'visitor' stamp: 'nice 1/9/2019 18:24'!
visitMethodNode: aMethodNode
	aMethodNode arguments size < 16 ifTrue: [ ^super visitMethodNode: aMethodNode ].
	^self visitMethodNodeWithManyArgs: aMethodNode! !

!IRTranslatorV2 methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:43'!
visit1ArrayForAllArgument: oneArray
	"Don't generate code, but remember the associated variables"
	
	tempVectorStack push: oneArray.! !

!IRTranslatorV2 methodsFor: '*Smallapack-OpalCompiler' stamp: 'nice 1/9/2019 18:59'!
visitPushArrayedArg: remoteTemp
	| tempIndex argumentRank tempVector |
	
	tempVector := tempVectorStack detect: [:each | each name = remoteTemp tempVectorName].
	
			"we may hardcode 0 because we know that the argument array is the first temp"
	tempIndex := self currentScope indexForVarNamed: remoteTemp tempVectorName.
	
	argumentRank := tempVector indexForVarNamed: remoteTemp name.

	gen pushRemoteTemp: argumentRank inVectorAt: tempIndex
! !

RFASTTranslatorForEffect subclass: #SLRFASTTranslatorForEffect
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLRFASTTranslatorForEffect commentStamp: 'nice 1/9/2019 21:12' prior: 0!
I am an auxiliary of Refective AST to IR translator for Smallapack Opal Compiler.

I generate instructions that pop resulting value from the execution  stack.

I am specialized for sending  messages with more than 15 arguments.!

!SLRFASTTranslatorForEffect methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:43'!
classForEffect
	^SLRFASTTranslatorForEffect ! !

!SLRFASTTranslatorForEffect methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:44'!
classForValue
	^SLRFASTTranslatorForValue! !

!SLRFASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:31'!
emitMessageNode: aMessageNode
	aMessageNode arguments size < 16 ifTrue: [ ^super emitMessageNode: aMessageNode ].
	^ self emitMessageNodeWithManyArgs: aMessageNode! !

!SLRFASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:29'!
emitMessageNodeWithManyArgs: aMessageNode
	super emitMessageNodeWithManyArgs: aMessageNode.
	methodBuilder popTop.! !

OpalCompiler subclass: #SLOpalCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLOpalCompiler commentStamp: 'nice 1/9/2019 21:55' prior: 0!
I am a variant of Opal Compiler used for Smallapack.

I am able to generate methods with more than 15 arguments, and send messages with more than 15 arguments.

My strategy is to pass ALL arguments into a single array argument when more than 15.
An alternative proposed by Eliot Miranda would be to pass 14 first arguments unchanged, and all excess arguments into an array in 15th-position.
But my purpose for Smallapack is to generate FFI calls, and for that, a single array of arguments is required, that explains why the single array for all arguments has my preference.

For this purpose, I am using a specialized compilation context.!

!SLOpalCompiler methodsFor: 'accessing' stamp: 'nice 1/9/2019 14:50'!
compilationContextClass
	^compilationContextClass ifNil: [ SLCompilationContext  ]! !

RFSemanticAnalyzer subclass: #SLRFSemanticAnalyzer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLRFSemanticAnalyzer commentStamp: 'nice 1/9/2019 21:23' prior: 0!
I am the Reflective semantic analyzer for Smallapack Opal Compiler.

I know how to deal with method arguments when they are passed thru a single array.!

!SLRFSemanticAnalyzer methodsFor: 'visitor' stamp: 'nice 1/9/2019 21:23'!
visitMethodNode: aMethodNode
	aMethodNode arguments size < 16 ifTrue: [ ^super visitMethodNode: aMethodNode ].
	^self visitMethodNodeWithManyArgs: aMethodNode! !

IRTempVector variableSubclass: #IR1ArrayForAllArgument
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!IR1ArrayForAllArgument commentStamp: 'nice 1/9/2019 21:03' prior: 0!
I represent a pseudo instruction used when passing all arguments into a single Array

This construct  is used for by-passing the 15-arguments limitation of bytecode (more exactly, the limitation of CompiledCode header).

I do not generate any code, because there isn't any code needed.
I am there only for mapping argument variable names to 0-based position into the array.!

!IR1ArrayForAllArgument methodsFor: 'visiting' stamp: 'nice 1/9/2019 18:41'!
accept: aVisitor
	^ aVisitor visit1ArrayForAllArgument: self! !

CompilationContext subclass: #SLCompilationContext
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLCompilationContext commentStamp: 'nice 1/9/2019 21:16' prior: 0!
I am a variant of Opal compilation context used for Smallapack.

I am able to generate methods with more than 15 arguments, and send messages with more than 15 arguments.

For this purpose, I am using specialized ancillary classes
- for semantic analysis of methods
- for AST to IR tanslation
- for IR to bytecode translation
!

!SLCompilationContext methodsFor: 'accessing' stamp: 'nice 1/9/2019 15:12'!
astTranslatorClass
	^ astTranslatorClass ifNil: [ astTranslatorClass := SLASTTranslator ]! !

!SLCompilationContext methodsFor: 'accessing' stamp: 'nice 1/9/2019 14:48'!
bytecodeGeneratorClass
	^ bytecodeGeneratorClass ifNil: [ bytecodeGeneratorClass := SLIRBytecodeGenerator ]! !

!SLCompilationContext methodsFor: 'accessing' stamp: 'nice 1/9/2019 16:00'!
semanticAnalyzerClass
	^ semanticAnalyzerClass ifNil: [ semanticAnalyzerClass := SLASTSemanticAnalyzer ]! !

OCASTTranslatorForEffect subclass: #SLASTTranslatorForEffect
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Smallapack-OpalCompiler'!

!SLASTTranslatorForEffect commentStamp: 'nice 1/9/2019 21:12' prior: 0!
I am an auxiliary of AST to IR translator for Smallapack Opal Compiler.

I generate instructions that pop resulting value from the execution  stack.

I am specialized for sending  messages with more than 15 arguments.!

!SLASTTranslatorForEffect methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:42'!
classForEffect
	^SLASTTranslatorForEffect ! !

!SLASTTranslatorForEffect methodsFor: 'initialization' stamp: 'nice 1/9/2019 14:44'!
classForValue
	^SLASTTranslatorForValue! !

!SLASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:31'!
emitMessageNode: aMessageNode
	aMessageNode arguments size < 16 ifTrue: [ ^super emitMessageNode: aMessageNode ].
	^ self emitMessageNodeWithManyArgs: aMessageNode! !

!SLASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'nice 1/9/2019 17:28'!
emitMessageNodeWithManyArgs: aMessageNode
	super emitMessageNodeWithManyArgs: aMessageNode.
	methodBuilder popTop.! !
